// ======================================================================= //
// Copyright (C)  1999, Linbox project
// Givaro / Athapascan-1
// Valence computation
// Time-stamp: <16 Apr 03 15:26:40 Jean-Guillaume.Dumas@imag.fr> 
// ======================================================================= //
#ifndef __LINBOX_VALENCE_C__
#define __LINBOX_VALENCE_C__

#include <math.h>
#include <sstream>

#include "givintprime.h"
#include "givintrns.h"

#include "LinBox/lin_rand.h"                      // Random Iterator
#include "LinBox/lin_spv_bb.h"                    // BB Wrapper for sparse vectors
#include "LinBox/lin_symmetrize_bbit.h"           // BB iterator
#include "LinBox/lin_massey.C"                // massey reccuring sequence solver

typedef SparseBlackBoxDom< GFqDomain > SPBB ;
    // Compute the rank of B^T B
    // with B == D1 A^T D2^2 A D1
typedef BB_Symmetrize_Container< SPBB > SzCBB;


struct New_Prime {
    void operator() (Random& generator, std::vector<IntPrimeDom::element>& vp) {
        IntPrimeDom I;
        IntPrimeDom::element P;
        std::vector<IntPrimeDom::element>::const_iterator vi;
        if (vp.size())
            do {
                unsigned long ss; generator(ss);
                I.nextprime(P, IntPrimeDom::element(MINLOWERSIZE+(ss%MINLOWERSIZE)));
                for (vi = vp.begin(); vi != vp.end(); ++vi)
                    if ( I.areEqual(*vi, P) ) break;
            } while ( vi != vp.end() ) ;
        else {
            unsigned long long ss; generator(ss);
            I.nextprime(P, IntPrimeDom::element(MINLOWERSIZE+(ss%MINLOWERSIZE)));
        }
        vp.push_back( P );
    }
};


struct One_Wiedemann {
    void operator()(IntPrimeDom::element& V, unsigned long long& rk,  char * Matrix_File_Name, Random& generator, unsigned long long p, unsigned long long Early_Term) {
        GFqDomain F(p);
        SPBB MF( F , Matrix_File_Name );
        SzCBB TF( &MF, generator);
        MasseyDom< SzCBB >  WD( &TF, Early_Term );
        GFqDomain::Rep v;
        WD.valence(v,rk);
        F.write(V, v);
    }
};



void seq_Integer_valence(char * Matrix_File_Name, Integer& Valence, unsigned long & rank, double OCB, double& wrong, double error = 0) {

    startActivity(GlobalCommentator, "Int Vale", "Integer Valence", LVL_NORMAL, INTERNAL_DESCRIPTION ); 


    
    Random generator;
    unsigned long long Early_Term = WIED_EARLYTERM_THRESHOLD;
    unsigned long long valBOUND = RNS_EARLYTERM_THRESHOLD;

    std::vector<IntPrimeDom::element> p;
    New_Prime() (generator, p);

    std::vector<IntPrimeDom::element> V(1);
    std::vector<unsigned long long> rk(1);

    One_Wiedemann() (V[0], rk[0], Matrix_File_Name, generator, (unsigned long long)p[0], Early_Term);
    if (rk[0] & 0x1) {
        if (V[0] > 0)
            V[0] -= p[0];
    } else
        if (V[0] < 0)
            V[0] +=  p[0];

    // For probability bounds, see :
    // [Dumas, Saunders, Villard. On efficient sparse integer matrix Smith normal form computations. JSC 32 (1/2), pp 71-99, 2001].

    double maxsize = (double)rk[0] * log(2.0*OCB) / log( (double)MINLOWERSIZE);
    double bad = (double)(rk[0]+1)*(double)(rk[0]+1)*log(OCB) / log( (double)MINLOWERSIZE) / 2.0;

    if ((error > 0) && (error < 1)) {
            valBOUND = 0; // no early termination in this case
            double logbad = GIVMAX( 1, -log( bad / NUMBEROFPRIMES ) );
            double maxerror = -log ( error ) / logbad;
            if (maxerror > maxsize) {
                activityReport(GlobalCommentator, "Taking high probability into account...", LVL_NORMAL, INTERNAL_DESCRIPTION); 
                maxsize = maxerror;
            }
    }

    unsigned long long NBLAUNCH = (unsigned long long)ceil( maxsize );
    { 
        std::ostringstream ost; ost << NBLAUNCH << " modular valences to be computed";
        std::string s = ost.str();
        activityReport(GlobalCommentator, s.c_str(), LVL_NORMAL, INTERNAL_DESCRIPTION);
    }
    
    { 
        std::ostringstream ost;
        ost << "P[0]: " << p[0]
            << ", V[0]: " << V[0]
            << ", rk[0]: " << rk[0];
        std::string s = ost.str();
        activityReport(GlobalCommentator, s.c_str(), LVL_UNIMP, INTERNAL_DESCRIPTION);
    }
    



    Valence = V[0];
    unsigned long long diff_bound = 0;

    for(unsigned long long i=1;i<NBLAUNCH;++i) {
        V.resize(V.size()+1);
        rk.resize(rk.size()+1);
        New_Prime() (generator, p);
        One_Wiedemann() (V[i], rk[i], Matrix_File_Name, generator, (unsigned long long)p[i], Early_Term);

        { 
            std::ostringstream ost;
            ost << "P[" << i << "]: " << p[i]
                << ", V[" << i << "]: " << V[i]
                << ", rk[" << i << "]: " << rk[i]
                << ", db: " << diff_bound;
            std::string s = ost.str();
            activityReport(GlobalCommentator, s.c_str(), LVL_UNIMP, INTERNAL_DESCRIPTION);
        }
           
        if (rk[i] == rk[0]) {
                IntRNSsystem<std::vector,std::allocator> RNs(p);
                IntRNSsystem<std::vector,std::allocator>::element V0;
                RNs.RnsToRing(V0, V);
                if (rk[i] & 0x1) {
                        if (V0 > 0)
                                V0 -= RNs.product();
                } else
                        if (V0 < 0)
                                V0 += RNs.product();
                if ((V0 == Valence) && (valBOUND > 0) ) if (++diff_bound > valBOUND) {
                        NBLAUNCH = i+1;
                        activityReport(GlobalCommentator, "RNS Early Termination", LVL_IMP, INTERNAL_DESCRIPTION);
                        break;
                }
                Valence = V0;
        }
    }

    
    {

        

        std::ostringstream ost;
        double wrong = pow( (double)bad/(double)NUMBEROFPRIMES, (double) NBLAUNCH );
        if (wrong < 1) {
            if (wrong <= 0) {
                ost << "Valence is " << Valence << ", correct with probability at least 1 - 1e" << (long long)(NBLAUNCH*log((double)bad/(double)NUMBEROFPRIMES)/log(10.0) );
            } else {
                ost << "Valence is " << Valence << ", correct with probability at least 1 - " << wrong;
            }
        } else {
            ost << "Valence is " << Valence << ", heuristic (probability of correctness unknown).";
        }
        std::string s = ost.str();
        activityReport(GlobalCommentator, s.c_str(), LVL_IMP, PARTIAL_RESULT);
    }
    



}

#endif 
